<?php
/**
 * svn.class.php -- PHP library for communicating with SVN server through 
 *                  the SVN protocol
 *
 * Copyright (C) 2009 Martin "Arcao" Sloup <arcao@arcao.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * SVN protocol specification: http://svn.collab.net/repos/svn/trunk/subversion/libsvn_ra_svn/protocol
 */

/**
 * Class for communicating with SVN server through the SVN protocol
 *
 * @author     Martin "Arcao" Sloup
 * @copyright  Copyright (C) 2009 Martin "Arcao" Sloup <arcao@arcao.com>
 * @package svn 
 */
class svn {
	const BUFFER_SIZE = 8192;
	const USER_AGENT = 'PHP SVN protocol class (svn.class.php/0.1)';
	const CONNECTION_TIMEOUT = 2; //in seconds
	const SVN_DEFAULT_PORT = 3690;
	const SVN_EMPTY = '( )';

	private $url;
	private $fp;
	
	/**
	 * SVN Class constructor. 
	 * @param string SVN url in format: svn://host[:port]/path	 
	 * @param string username (for anonymous login use 'anonymous' or empty string)
	 * @param string password
	 */
	public function __construct($url, $username = 'anonymous', $password = '') {
		$this->url = parse_url($url);
		
		$port = (isset($this->url['port'])) ? (int) $this->url['port'] : self::SVN_DEFAULT_PORT;
		
		$this->fp = @fsockopen("tcp://".$this->url['host'], $port, $errNo, $errStr, self::CONNECTION_TIMEOUT);
		if (!$this->fp) { 
			// error
			throw new ConnectionException($errStr, $errNo); 
			return; 
		}
		
		stream_set_blocking($this->fp, 0);
		
		self::checkError($this->in());
		
		$url = "svn://{$this->url['host']}{$this->url['path']}";
		
		$this->out(Array(
			2, //protocol ver
			Array('edit-pipeline', 'svndiff1', 'absent-entries', 'depth', 'mergeinfo', 'log-revprops'), //caps
			self::_s($url), //url
			self::_s(self::USER_AGENT), //user agent
			self::SVN_EMPTY // empty
		));
		$response = $this->in();
		self::checkError($response);
		
		$anonymousLogin = (!$username || strtolower($username) == 'anonymous');
		
		if ($anonymousLogin && strpos(implode(' ', $response[0][1][0]), 'ANONYMOUS') === false)
			throw new LoginMethodException("Anonymous login is not allowed.");
		
		if (!$anonymousLogin && strpos(implode(' ', $response[0][1][0]), 'CRAM-MD5') === false)
			throw new LoginMethodException("CRAM-MD5 login is not supported.");
		
		if ($anonymousLogin) {
			$this->out(self::call('ANONYMOUS', 
				self::_s(base64_encode('anonymous@'.$_SERVER['REMOTE_HOST']))	//anonymous password
			));
			$response = $this->in();

			if ($response[0][0] != 'success')
				throw new LoginException($response[0][1][0]);

		} else {
			$this->out(self::call('CRAM-MD5'));
			$response = $this->in();
						
			if ($response[0][0] != 'step')
				throw new LoginMethodException("CRAM-MD5 login step error.");
			
			$key = $response[0][1][0];
			$this->out(self::_s($username.' '.self::hmacMd5($password, $key)));
			
			if ($response[0][0] != 'success')
				throw new LoginException($response[0][1][0]);
		}
	}
	
	/**
	 * Returns list of items in directory AS an array. If direcotory is not exist, returns empty array.  
	 * @param string path
	 * @param int revision if is not specified the last revision is used
	 * @return arrray array of associated array with these keys: name, kind (file, dir), size (in bytes), has-props, cereated-rev, time (unix time), last-author  
	 */
	public function getDir($path, $rev = -1) {
		if ($rev == -1)
			$rev = $this->getLastRev();
		
		if (!$this->checkPath($path, $rev))
			return Array();
		
		$this->out(self::call('get-dir', 
			self::_s($path), //path
			Array($rev), //revision
			false, //want-props
			true, //want-content
			Array('kind', 'size', 'has-props', 'created-rev', 'time', 'last-author') //je uplne jedno, zda se to specifikuje nebo ne, vzdy to prijde v tomto poradi
		));
		$response = $this->in();
			
		if ($response[1][0] != 'success')
			return Array();
		
		$lines = $response[1][1][2];
		
		$dir = Array();
		foreach($lines as $item) {
			$dir[] = Array(
				'name' => $item[0],
				'kind' => $item[1],
				'size' => (int) $item[2],
				'has-props' => (bool) $item[3],
				'created-rev' => (int) $item[4],
				'time' => strtotime($item[5][0]),
				'last-autor' => $item[6][0]
			);
		}
		
		return $dir;
	}
	
	/**
	 * Tests if the path exist and what a kind of path is it (file or directory).   
	 * @param string path
	 * @param int revision if is not specified the last revision is used
	 * @return string|bool 'dir' or 'file' if path exist, otherwise returns false  
	 */	
	public function checkPath($path, $rev = -1) {
		if ($rev == -1)
			$rev = $this->getLastRev();
			
		$this->out(self::call('check-path',
			self::_s($path), //path
			Array($rev) //revision
		));
		$response = $this->in();

		
		if ($response[1][1][0] != 'dir' && $response[1][1][0] != 'file')
			return false;
			
		return $response[1][1][0];
	}

	/**
	 * Returns the last revision in repository   
	 * @return int last revision as an integer or -1 if some error happend   
	 */			
	public function getLastRev() {
		$this->out(self::call('get-latest-rev'));
		$response = $this->in();
		
		if ($response[1][0] != 'success')
			return -1;
			
		return $response[1][1][0];
	}
	
	/**
	 * Retrieve file from repository.
	 *
	 * If the second parameter is a resource (eg. open with fopen) function write the output to the resource:
	 * <code>
	 * $fp = fopen('file.txt', 'w');
	 * $svn->getFile('/path/file.txt', $fp);
	 * fclose($fp); 	 	 
	 * </code>
	 *	 	 
	 * @param string filename
	 * @param resource|bool if is a resource, write a file content to the resource, othervise returns the file content.
	 * @param int revision if is not specified the last revision is used	 	 	    
	 * @return string|bool if the second perameter is resource returns true for successful retrieving, othervise returns file content
	 */
	public function getFile($filename, $fp_out = false, $rev = -1) {
		if ($rev == -1)
			$rev = $this->getLastRev();
		
		if (!$this->checkPath($filename, $rev))
			return false;		
	
		$this->out(self::call('get-file',
			self::_s($filename), //filename
			Array($rev), //revision
			false, //want-props
			true //want-contents
		));
		
		return $this->inData($fp_out);
	}
	
	/**
	 * Returns informarmation about file / dir specified in $path parameter. If path doesn't exist, returns empty array.   
	 * @param string path
	 * @param int revision if is not specified the last revision is used
	 * @return arrray associated array with these keys:  kind (file, dir), size (in bytes), has-props, cereated-rev, time (unix time), last-author  
	 */
	public function stat($path, $rev = -1) {
		if ($rev == -1)
			$rev = $this->getLastRev();
			
		$this->out(self::call('stat',
			self::_s($path), //path
			Array($rev) //revision
		));
		$response = $this->in();
		
		if ($response[1][0] != 'success' || Count($response[1][1][0]) == 0)
			return Array();
		
		$item = $response[1][1][0][0];
		
		return Array(
			'kind' => $item[0],
			'size' => (int) $item[1],
			'has-props' => (bool) $item[2],
			'created-rev' => (int) $item[3],
			'time' => strtotime($item[4][0]),
			'last-autor' => $item[5][0]
		);		
	}
	
	/**
	 * Close a connectioun to the SVN server.
	 */	 	
	public function close() {
		fclose($this->fp);
	}

	//------------------- private functions --------------------------------------
	
	private function in() {
		$data = '';
		while (($buf = fread($this->fp, self::BUFFER_SIZE)) != '' || strlen($data) <2 || substr($data, -2, 2) != ') ')
			$data.= $buf;
		
		//echo "-> $data\n";
		//print_r(self::parseIn($data));
		//flush();
		
		return self::parseIn($data);
	}
	
	private function out($data) {
		if (is_array($data))
			$data = self::toSvnLine($data);
			
		//echo "<- $data\n";
		//flush();
		
		fwrite($this->fp, $data.' ');
	}
	
	private function inData($fp_out) {
		$ret = ''; //pokud $fp_out je false, vracime obsah souboru misto vypisu do streamu
		
		$buffer = '';
		$status = 0; //0=begin, 1=in first success response, 2=out of success response, 3=file hash response, 4=in data response, 5=in end response, 6=end
		$level = 0; //uroven zanoreni zavorek
		$pos = 0;
		
		while ($status < 6) {
			//na zacatku cyklu precteme vzdy prichozi data
			$buffer.= fread($this->fp, self::BUFFER_SIZE);
			//pokud neni nic bufferu skocime na zacatek cyklu
			if (strlen($buffer) == 0)
				continue;
			
			//odstraneni nepotrebnych dat z bufferu
			if ($pos > 4096) {
				$buffer = substr($buffer, $pos);
				$pos = 0;
			}
			
			//nacteme znak z bufferu, detekujeme zavorky
			$ch = $buffer[$pos];
			if ($ch == '(') {
				//sledujeme jen zanoreni
				$pos+=2;
				$level++;
				if ($level == 1)
					$status++;	
			} elseif ($ch == ')') {
				//sledujeme jen zanoreni
				$pos+=2;
				$level--;
				if ($level == 0)
					$status++;
			} elseif (is_numeric($ch)) {
				//jeli znak cislo, muzou to byt dve moznosti, bud je to opravdu jen cislo, nebo se jedna o text ve tvaru <delka>:<text>
				$next1 = strpos($buffer, ':', $pos);
				$next2 = strpos($buffer, ' ', $pos);
				
				if ($next1 === false && $next2 === false)
					continue; //malo dat, pokusime se precist dalsi data
				
				if ($next1 === false || ($next2 !== false && ($next2 < $next1))) {
					$next = $next2;
					if ($next === false)
						continue; //malo dat, pokusime se precist dalsi data
				
					//cisla nas nezajimaji, preskakujeme
					$pos = $next + 1;
				} else {
					$next = $next1;
					//zjistime delku textu
					$len = substr($buffer, $pos, $next - $pos);
					
					if (strlen($buffer) < $next + $len + 2)
						continue;	//malo dat, pokusime se precist dalsi data
					
					//posuneme se na zacatek textu
					$pos = $next + 1;
					
					//nejsme-li v bloku s daty souboru, obsah nas nezajima
					if ($status == 4) {
						$data = substr($buffer, $pos, $len);
						if (is_resource($fp_out)) {
							fwrite($fp_out, $data);
						} else {
							$ret.= $data;
						}
					}
					
					$pos+= $len + 1;
				}
			} else {
				//jedna se o "textovou konstantu"
				$next = strpos($buffer, ' ', $pos);
				if ($next === false)
					continue; //malo dat, pokusime se precist dalsi data
				
				if ($status == 3) {
					$test = substr($buffer, $pos, $next - $pos);
					if ($test == 'failure') {
						$response = self::parseIn(substr($buffer, $pos));
						throw new FileException($response[1][0][1], $response[1][0][0]);
					}
				}
				//konstanty ve forme textu nas nezajimaji
				$pos = $next + 1;
			}
		}
	
		return ($fp_out === false) ? $ret : true;
	}
	
	//------------------- private static functions -------------------------------
	
	private static function checkError($response) {
		if (!is_array($response))
			throw new Exception('Parameter must be an array.');
			
		if ($response[0][0] != 'success') {
			throw new Exception($response[0][1][0]);
		}			
	}
	
	private static function _s($string) {
		return strlen($string).':'.$string;
	}
	
	private static function parseIn($buffer) {
		$pos = 0;
		return self::_parseIn($buffer, $pos);
	}
	
	private static function _parseIn(&$buffer, &$pos) {
		$ret = Array(); 
		while ($pos < strlen($buffer)) {
			//odstraneni nepotrebnych dat z bufferu
			if ($pos > 4096) {
				$buffer = substr($buffer, $pos);
				$pos = 0;
			}
		
			$ch = $buffer[$pos];
			//zavorky znamenaji zanoreni
			if ($ch == '(') {
				$pos+=2;
				$ret[] = self::_parseIn($buffer, $pos);
			} elseif ($ch == ')') {
				$pos+=2;
				break;
			} elseif (is_numeric($ch)) {
				//jeli znak cislo, muzou to byt dve moznosti, bud je to opravdu jen cislo, nebo se jedna o text ve tvaru <delka>:<text>
				$next1 = strpos($buffer, ':', $pos);
				$next2 = strpos($buffer, ' ', $pos);
				
				if ($next1 === false || $next2 < $next1) {
					//je to cislo
					$next = $next2;
					if ($next === false)
						throw new Exception('ListItem terminator not found on position '.$pos.'.');
				
					$ret[] = substr($buffer, $pos, $next - $pos);
					$pos = $next + 1;
				} else {
					//je to text
					$next = $next1;
					$len = substr($buffer, $pos, $next - $pos);
					$pos = $next + 1;
					
					$ret[] = substr($buffer, $pos, $len);
					$pos+= $len + 1;
				}
			} else {
				//jedna se o "textovou konstantu"
				$next = strpos($buffer, ' ', $pos);
				if ($next === false)
					throw new Exception('ListItem terminator not found on position '.$pos.'.');
				
				$ret[] = substr($buffer, $pos, $next - $pos);
				$pos = $next + 1;
			}
		}
		
		return $ret;
	}
	
	private static function hmacMd5($password, $key) {
		$password = strlen($password) < 64 ? str_pad($password, 64, "\0") : substr($password, 0, 64);
		return md5((str_repeat("\x5c", 64) ^ $password) . pack("H32", md5( (str_repeat("\x36", 64)^$password).$key)));
  }
  
  private static function toSvnLine($arr) {
  	$ret = '( ';
  	foreach($arr AS $item) {
  		if (is_array($item)) {
  			$ret.=self::toSvnLine($item);
  		} elseif ($item===true) {
  			$ret.='true ';
  		} elseif ($item===false) {
  			$ret.='false ';
			} else {
  			$ret.=$item.' ';
			}
		}
		$ret.=') ';
		
		return $ret;
	}
	
	private static function call() {
		if (func_num_args() == 0)
			throw new Exception('Command parameter must be specified.');
		
		$args = func_get_args();
		
		if (func_num_args() == 1) {
			$param = self::SVN_EMPTY;
		} else {
			$param = array_slice($args, 1);
		}
		 
		return Array($args[0], $param);
	}
}

//------------------- Exceptions -----------------------------------------------
class ConnectionException extends Exception {};
class LoginException extends Exception {};
class LoginMethodException extends LoginException {};
class FileException extends Exception {};